import os
import re
import tqdm
import shutil
import asyncio
import aiohttp
import aiofiles
from . import utils
from bs4 import BeautifulSoup, SoupStrainer
from multiprocessing import Process, Queue
from typing import List


def _korean_characters_ratio(text: str) -> float:
    korean_characters = 0
    for c in text:
        if ord("가") <= ord(c) and ord(c) <= ord('힣'):
            korean_characters += 1
    return korean_characters / len(text)


async def _get_max_nav_pages(sess: aiohttp.ClientSession,
                             category: int,
                             date: str,
                             max_page: int = 100) -> int:
    nav_url = ('https://news.naver.com/main/list.nhn?mode=LSD&mid=shm'
               '&sid1={category}&date={date}&page={page}'
               .format(category=category, date=date, page=max_page))

    # Read navigation page and extract current page from container.
    async with sess.request('GET', nav_url) as resp:
        document = await resp.text()

    document = document[document.find('<div class="paging">'):]
    document = document[:document.find('</div>')]

    return int(re.search(r'<strong>(.*?)</strong>', document).group(1))


async def _get_article_urls_from_nav_page(sess: aiohttp.ClientSession,
                                          category: int,
                                          date: str,
                                          page: int) -> List[str]:
    nav_url = ('https://news.naver.com/main/list.nhn?mode=LSD&mid=shm'
               '&sid1={category}&date={date}&page={page}'
               .format(category=category, date=date, page=page))

    # Read navigation page and extract article links.
    async with sess.request('GET', nav_url) as resp:
        document = await resp.text()

    document = document[document.find('<ul class="type06_headline">'):]

    # Extract article url containers.
    list1 = document[:document.find('</ul>')]
    list2 = document[document.find('</ul>') + 5:]
    list2 = list2[:list2.find('</ul>')]

    document = list1 + list2

    # Extract all article urls from their containers.
    article_urls = []
    while '<dt>' in document:
        document = document[document.find('<dt>'):]
        container = document[:document.find('</dt>')]

        if not container.strip():
            continue

        article_urls.append(
            re.search(r'<a href="(.*?)"', container).group(1))

        document = document[document.find('</dt>'):]

    return article_urls


async def _get_article_content(sess: aiohttp.ClientSession,
                               article_url: str) -> str:
    # Use `SoupStrainer` to improve performance.
    async with sess.request('GET', article_url) as resp:
        document = await resp.text()

    strainer = SoupStrainer('div', attrs={'id': 'articleBodyContents'})
    document = BeautifulSoup(document, 'lxml', parse_only=strainer)
    content = document.find('div')

    # Skip if there is no article content in the page.
    if content is None:
        return ''

    # Remove unnecessary tags.
    for child in content.find_all():
        child.clear()

    return ' '.join(content.get_text().split())


def _collect_article_urls_worker(queue: Queue,
                                 category_list: List[str],
                                 date_list: List[str],
                                 max_page: int = 100,
                                 max_tasks: int = 50):
    async def main_fn():
        @utils.notifiable
        async def async_fn(sess, category, date):
            article_urls = []

            # Get maximum number of pages in navigation.
            pages = await _get_max_nav_pages(sess, category, date, max_page)

            # Collect all article urls from the given page index.
            for page in range(1, pages + 1):
                article_urls += await _get_article_urls_from_nav_page(
                    sess, category, date, page)

            return article_urls

        ctx = utils.Context(max_tasks=max_tasks)
        async with aiohttp.ClientSession() as sess:
            article_urls_stack, errors = await ctx.run((
                async_fn(sess, category, date,
                         callback=lambda: queue.put(None))
                for category in category_list
                for date in date_list))

        # Flatten the collected article urls and return it by putting to the
        # queue.
        queue.put(sum(article_urls_stack, []))

    # Execute async main function.
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_fn())


def _crawl_articles_worker(queue: Queue,
                           output_file: str,
                           article_urls: List[str],
                           max_tasks: int = 50):
    async def main_fn():
        @utils.notifiable
        async def async_fn(sess, article_url, fp):
            # Get news article content and save it to the output file.
            content = await _get_article_content(sess, article_url)
            if _korean_characters_ratio(content) > 0.5:
                await fp.write(content + '\n')

        ctx = utils.Context(max_tasks=max_tasks)
        async with aiohttp.ClientSession() as sess:
            async with aiofiles.open(output_file, 'w', encoding='utf-8') as fp:
                _, errors = await ctx.run((
                    async_fn(sess, article_url, fp,
                             callback=lambda: queue.put(True))
                    for article_url in article_urls))

        # Notify that current process is finished.
        queue.put(None)

    # Execute async main function.
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_fn())


def start_crawling_articles(output_file: str,
                            temporary: str,
                            num_cores: int,
                            category_list: List[str],
                            start_date: str,
                            end_date: str,
                            date_step: int,
                            max_page: int = 100,
                            max_tasks: int = 50):
    date_list = utils.drange(start_date, end_date, date_step)
    total_search = len(date_list) * len(category_list)

    # Prepare multi-processing.
    workers = []
    queue = Queue()
    date_list_chunks = utils.split_list(date_list, chunks=num_cores)

    for i in range(num_cores):
        w = Process(target=_collect_article_urls_worker,
                    args=(queue,
                          category_list,
                          date_list_chunks[i],
                          max_page,
                          max_tasks))
        w.daemon = True
        w.start()

        workers.append(w)

    # Gather article urls from processes.
    article_urls = []
    exit_processes = 0
    tqdm_iter = tqdm.trange(total_search, desc='[*] collect article urls')
    while True:
        batch_article_urls = queue.get()

        if batch_article_urls is None:
            tqdm_iter.update()
        else:
            article_urls += batch_article_urls
            exit_processes += 1

        # Exit for waiting processes.
        if exit_processes == num_cores:
            break
    tqdm_iter.close()

    print(f'[*] successfully collecting article urls.'
          f' total articles: {len(article_urls)}')

    # Prepare multi-processing.
    workers = []
    queue = Queue()

    # Create temporary files and split articles into chunks.
    crawled_files = utils.random_filenames(temporary, num_cores)
    article_list_chunks = utils.split_list(article_urls, chunks=num_cores)

    for i in range(num_cores):
        w = Process(target=_crawl_articles_worker,
                    args=(queue,
                          crawled_files[i],
                          article_list_chunks[i],
                          max_tasks))
        w.daemon = True
        w.start()

        workers.append(w)

    # Show crawling progress.
    exit_processes = 0
    for _ in tqdm.trange(len(article_urls), desc='[*] crawl articles'):
        if queue.get() is None:
            exit_processes += 1

        # Exit for waiting processes.
        if exit_processes == num_cores:
            break

    # Merge temporary files into `output_file`.
    print(f'[*] finish crawling articles. merge chunks into [{output_file}].')
    with open(output_file, 'wb') as dst:
        for name in crawled_files:
            with open(name, 'rb') as src:
                shutil.copyfileobj(src, dst)

    # Remove temporary files.
    for name in crawled_files:
        os.remove(name)



import os
import random
import string
import asyncio
import functools
from datetime import datetime, timedelta
from typing import List, Iterable, Any, Callable


def notifiable(func):
    @functools.wraps(func)
    async def wrapper(*args, callback: Callable = None, **kwargs):
        try:
            result = await func(*args, **kwargs)
        finally:
            # If callback is given, then call callback function to notify that
            # the given function is finished.
            if callback is not None:
                callback()
        return result
    return wrapper


def drange(start: str, end: str, step: int = 1) -> List[str]:
    start_date = datetime.strptime(start, '%Y%m%d')
    end_date = datetime.strptime(end, '%Y%m%d')

    iters = (end_date - start_date).days // step
    return [(start_date + timedelta(days=d * step)).strftime('%Y%m%d')
            for d in range(iters + 1)]


def split_list(seq: List, chunks: int) -> List[List]:
    chunk_list = []
    chunk_size = len(seq) // chunks

    for i in range(chunks - 1):
        chunk_list.append(seq[i * chunk_size:(i + 1) * chunk_size])
    chunk_list.append(seq[(chunks - 1) * chunk_size:])

    return chunk_list


def random_filename(parent: str) -> str:
    return os.path.join(parent, ''.join(random.choices(
        string.digits + string.ascii_lowercase, k=16)))


def random_filenames(parent: str, n: int) -> List[str]:
    return [random_filename(parent) for _ in range(n)]


class Context(object):
    def __init__(self, max_tasks: int):
        self.max_tasks = max_tasks

    async def run(self, aws: Iterable) -> Any:
        tasks, results, errors = set(), [], 0
        for a in aws:
            if len(tasks) >= self.max_tasks:
                # If pool is full, wait for the first-completed task.
                done, tasks = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED)

                # Update results or exceptions.
                for d in done:
                    try:
                        r = d.result()
                        if r is not None:
                            results.append(r)
                    except Exception:
                        errors += 1

            # Add new task and put into the pool.
            tasks.add(asyncio.ensure_future(a))

        # After adding all tasks to the pool, wait for the rest tasks.
        done, _ = await asyncio.wait(tasks)

        # Update results or exceptions.
        for d in done:
            try:
                r = d.result()
                if r is not None:
                    results.append(r)
            except Exception:
                errors += 1

        return results, errors



import os
import re
import kss
import shutil
from . import utils
from multiprocessing import Process, Queue
from typing import Dict, Any


_re_special_symbols = re.compile(r'[\/|*^@\#▲▶◆◀■【】\\\=]')


def _normalize_quotes(text: str) -> str:
    text = text.replace('\u0060', '\'')
    text = text.replace('\u00b4', '\'')
    text = text.replace('\u2018', '\'')
    text = text.replace('\u2019', '\'')
    text = text.replace('\u201c', '\"')
    text = text.replace('\u201d', '\"')
    return text


def _clean_article_content(content: str) -> str:
    content = _re_special_symbols.sub(' ', content)
    content = content[content.find('다.') + 2:content.rfind('다.')]
    content = content[:content.rfind('다.') + 2]
    return ' '.join(content.split())


def _clean_articles_worker(output_file: str, queue: Queue):
    with open(output_file, 'w', encoding='utf-8') as fp:
        while True:
            content = queue.get()

            # Terminate the process if queue returns `None`.
            if content is None:
                break

            # Write cleaned article content to output file.
            fp.write(_clean_article_content(content) + '\n\n')


def _tokenize_sentences_worker(input_file: str, output_file: str,
                               min_len: int, max_len: int,
                               split_sent: bool = True):
    with open(input_file, 'r', encoding='utf-8') as src, \
            open(output_file, 'w', encoding='utf-8') as dst:
        total_lines = ''
        for line in src:
            if not line.strip():
                # Write the rest sentences.
                if not split_sent and len(total_lines.strip()) > min_len:
                    dst.write(_normalize_quotes(total_lines.strip()) + '\n')
                total_lines = ''
                continue

            for s in kss.split_sentences(line):
                s = s.strip()

                if split_sent:
                    if len(s) > min_len and len(s) < max_len:
                        dst.write(_normalize_quotes(s) + '\n')
                else:
                    if len(total_lines) + len(s) > max_len:
                        dst.write(_normalize_quotes(total_lines.strip())
                                  + '\n')
                        total_lines = ''
                    total_lines += s + ' '


def _process_canrevan_article_contents(input_file: str,
                                       output_file: str,
                                       temporary: str,
                                       args: Dict[str, Any]):
    # Prepare cleaning articles in parallel.
    workers = []
    queue = Queue(maxsize=50 * args['num-cores'])
    cleaned_files = utils.random_filenames(temporary, args['num-cores'])

    # Start processes.
    for i in range(args['num-cores']):
        w = Process(target=_clean_articles_worker,
                    args=(cleaned_files[i], queue))
        w.daemon = True
        w.start()

        workers.append(w)

    # Read file and send to each process.
    with open(input_file, 'r', encoding='utf-8') as fp:
        for line in fp:
            queue.put(line)

    # Send terminate message to each process.
    for _ in range(args['num-cores']):
        queue.put(None)

    for w in workers:
        w.join()

    # Prepare tokenizing sentences in parallel.
    workers = []
    tokenized_files = utils.random_filenames(temporary, args['num-cores'])

    # Start processes.
    for i in range(args['num-cores']):
        w = Process(target=_tokenize_sentences_worker,
                    args=(cleaned_files[i],
                          tokenized_files[i],
                          args['min-length'],
                          args['max-length'],
                          args['split-sent'] == 'true'))
        w.daemon = True
        w.start()

        workers.append(w)

    # Wait for terminating.
    for w in workers:
        w.join()

    # Remove temporary files.
    for name in cleaned_files:
        os.remove(name)

    # Merge temporary files into output file.
    with open(output_file, 'wb') as dst:
        for name in tokenized_files:
            with open(name, 'rb') as src:
                shutil.copyfileobj(src, dst)

    # Remove temporary files.
    for name in tokenized_files:
        os.remove(name)


__extension__ = {
    'name': 'canrevan extension',
    'version': '1.0',
    'description': 'clear crawled article contents.',
    'author': 'canrevan',
    'main': _process_canrevan_article_contents,
    'arguments': {
        'num-cores': {'type': int, 'default': 1},
        'min-length': {'type': int, 'default': 50},
        'max-length': {'type': int, 'default': 1000},
        'split-sent': {'type': str, 'default': 'true'},
    }
}
